"""
MIT License

Copyright (c) 2017 PADAS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__version__ = "cc3a89651c4c652ba0a8bbd4fed1117becc159fb"

from datetime import datetime, timedelta
from urllib3.util.retry import Retry
import pytz
import logging
import re
import concurrent.futures
import math
import requests
from requests.adapters import HTTPAdapter
import json

version_string = "1.0.40"


def linkify(url, params):
    p = ["=".join((str(x), str(y))) for x, y in params.items()]
    p = "&".join(p)
    return "?".join((url, p))


def split_link(url):
    url, qs = url.split("?")
    params = dict([p.split("=") for p in qs.split("&")])
    return (url, params)


class DasClient(object):
    """
    DasClient provides basic access to a DAS API. It requires the coordinates of a DAS API service as well
    as valid credentials for a user.

    The boiler-plate code handles authentication, so you don't have to think about Oauth2 or refresh tokens.

    As of May 12, 2017 it includes just a basic set of functions to access Subject data and to post observations.

    June 6, 2017: Added methods to add a photo or document to an Event.

    """

    def __init__(self, **kwargs):
        """
        Initialize a DasClient instance.

        :param username: DAS username
        :param password: DAS password
        :param service_root: The root of the DAS API (Ex. https://demo.pamdas.org/api/v1.0)
        :param token_url: The auth token url for DAS (Ex. https://demo.pamdas.org/oauth2/token)
        :param provider_key: provider-key for posting observation data (Ex. xyz_provider)
        :param client_id: Auth client ID (Ex. das_web_client)
        """

        self.auth = None
        self.auth_expires = pytz.utc.localize(datetime.min)
        self._http_session = None
        self.max_retries = kwargs.get("max_http_retries", 5)

        self.service_root = kwargs.get("service_root")
        self.client_id = kwargs.get("client_id")
        self.provider_key = kwargs.get("provider_key")

        self.token_url = kwargs.get("token_url")
        self.username = kwargs.get("username")
        self.password = kwargs.get("password")
        self.realtime_url = kwargs.get("realtime_url")

        if kwargs.get("token"):
            self.token = kwargs.get("token")
            self.auth = dict(token_type="Bearer", access_token=kwargs.get("token"))
            self.auth_expires = datetime(2099, 1, 1, tzinfo=pytz.utc)

        self.user_agent = "das-client/{}".format(version_string)

        self.logger = logging.getLogger(self.__class__.__name__)

        self._http_session = requests.Session()
        retries = Retry(total=5, backoff_factor=1.5, status_forcelist=[502])
        self._http_session.mount("http", HTTPAdapter(max_retries=retries))
        self._http_session.mount("https", HTTPAdapter(max_retries=retries))

    def _auth_is_valid(self):
        return self.auth_expires > pytz.utc.localize(datetime.utcnow())

    def auth_headers(self):

        if self.auth:
            if not self._auth_is_valid():
                if not self.refresh_token():
                    if not self.login():
                        raise DasClientException("Login failed.")
        else:
            if not self.login():
                raise DasClientException("Login failed.")

        return {
            "Authorization": "{} {}".format(
                self.auth["token_type"], self.auth["access_token"]
            ),
            "Accept-Type": "application/json",
        }

    def refresh_token(self):
        payload = {
            "grant_type": "refresh_token",
            "refresh_token": self.auth["refresh_token"],
            "client_id": self.client_id,
        }
        return self._token_request(payload)

    def login(self):

        payload = {
            "grant_type": "password",
            "username": self.username,
            "password": self.password,
            "client_id": self.client_id,
        }
        return self._token_request(payload)

    def _token_request(self, payload):

        response = requests.post(self.token_url, data=payload)
        if response.ok:
            self.auth = json.loads(response.text)
            expires_in = int(self.auth["expires_in"]) - 5 * 60
            self.auth_expires = pytz.utc.localize(datetime.utcnow()) + timedelta(
                seconds=expires_in
            )
            return True

        self.auth = None
        self.auth_expires = pytz.utc.localize(datetime.min)
        return False

    def _das_url(self, path):
        return "/".join((self.service_root, path))

    def _get(self, path, stream=False, **kwargs):
        headers = {"User-Agent": self.user_agent}

        headers.update(self.auth_headers())
        if not path.startswith("http"):
            path = self._das_url(path)

        response = None
        if self._http_session:
            response = self._http_session.get(
                path, headers=headers, params=kwargs.get("params"), stream=stream
            )
        else:
            response = requests.get(
                path, headers=headers, params=kwargs.get("params"), stream=stream
            )

        if response.ok:
            if kwargs.get("return_response", False):
                return response
            data = json.loads(response.text)
            if "metadata" in data:
                return data["metadata"]
            return data["data"]

        if response.status_code == 404:  # not found
            self.logger.error(f"404 when calling {path}")
            raise DasClientNotFound()

        if response.status_code == 403:  # forbidden
            try:
                _ = json.loads(response.text)
                reason = _["status"]["detail"]
            except:
                reason = "unknown reason"
            raise DasClientPermissionDenied(reason)

        self.logger.debug("Fail: " + response.text)
        raise DasClientException(
            f"Failed to call DAS web service. {response.status_code} {response.text}"
        )

    def _call(self, path, payload, method, params=None):
        headers = {"Content-Type": "application/json", "User-Agent": self.user_agent}
        headers.update(self.auth_headers())

        def time_converter(t):
            if isinstance(t, datetime):
                return t.isoformat()

        body = json.dumps(payload, default=time_converter)

        fmap = None
        if self._http_session:
            fmap = {"POST": self._http_session.post, "PATCH": self._http_session.patch}
        else:
            fmap = {"POST": requests.post, "PATCH": requests.patch}
        try:
            fn = fmap[method]
        except KeyError:
            self.logger.error("method must be one of...")
        else:
            response = fn(
                self._das_url(path), data=body, headers=headers, params=params
            )

        if response and response.ok:
            res_json = response.json()
            if "data" in res_json:
                return res_json["data"]
            else:
                return res_json

        if response.status_code == 404:  # not found
            self.logger.error(f"Could not load {path}")
            raise DasClientNotFound()

        try:
            _ = json.loads(response.text)
            reason = _["status"]["detail"]
        except:
            reason = "unknown reason"

        if response.status_code == 403:  # forbidden
            raise DasClientPermissionDenied(reason)

        if (
            response.status_code == 504 or response.status_code == 502
        ):  # gateway timeout or bad gateway
            self.logger.error(
                f"ER service unavailable",
                extra=dict(
                    provider_key=self.provider_key,
                    service=self.service_root,
                    path=path,
                    status_code=response.status_code,
                    reason=reason,
                    text=response.text,
                ),
            )
            raise DasClientServiceUnavailable(f"ER service unavailable")

        self.logger.error(
            f"ER returned bad response",
            extra=dict(
                provider_key=self.provider_key,
                service=self.service_root,
                path=path,
                status_code=response.status_code,
                reason=reason,
                text=response.text,
            ),
        )
        message = f"provider_key: {self.provider_key}, service: {self.service_root}, path: {path},\n\t {response.status_code} from ER. Message: {reason} {response.text}"
        raise DasClientException(f"Failed to {fn} to DAS web service. {message}")

    def _post(self, path, payload, params={}):
        return self._call(path, payload, "POST", params)

    def _patch(self, path, payload, params={}):
        return self._call(path, payload, "PATCH", params)

    def add_event_to_incident(self, event_id, incident_id):

        params = {"to_event_id": event_id, "type": "contains"}

        result = self._post("activity/event/" + incident_id + "/relationships", params)

    def remove_event_from_incident(
        self, event_id, incident_id, relationship_type="contains"
    ):
        result = self._delete(
            f"activity/event/{incident_id}/relationship/{relationship_type}/{event_id}/"
        )

    def _delete(self, path):

        headers = {"User-Agent": self.user_agent}
        headers.update(self.auth_headers())

        resonse = None
        if self._http_session:
            response = self._http_session.delete(self._das_url(path), headers=headers)
        else:
            response = requests.delete(self._das_url(path), headers=headers)
        if response.ok:
            return True

        if response.status_code == 404:  # not found
            self.logger.error(f"404 when calling {path}")
            raise DasClientNotFound()

        if response.status_code == 403:  # forbidden
            try:
                _ = json.loads(response.text)
                reason = _["status"]["detail"]
            except:
                reason = "unknown reason"
            raise DasClientPermissionDenied(reason)

        raise DasClientException(
            f"Failed to delete: {response.status_code} {response.text}"
        )

    def delete_event(self, event_id):
        self._delete("activity/event/" + event_id + "/")

    def delete_source(self, source_id):
        self._delete("source/" + source_id + "/")

    def delete_subject(self, subject_id):
        self._delete("subject/" + subject_id + "/")

    def delete_message(self, message_id):
        self._delete("messages/" + message_id + "/")

    def delete_patrol(self, patrol_id):
        self._delete("activity/patrols/" + patrol_id + "/")

    def _post_form(self, path, body=None, files=None):

        headers = {"User-Agent": self.user_agent}
        headers.update(self.auth_headers())

        body = body or {}
        response = requests.post(
            self._das_url(path), data=body, headers=headers, files=files
        )
        if response and response.ok:
            return json.loads(response.text)["data"]

        if response.status_code == 404:  # not found
            raise DasClientNotFound()

        if response.status_code == 403:  # forbidden
            try:
                _ = json.loads(response.text)
                reason = _["status"]["detail"]
            except:
                reason = "unknown reason"
            raise DasClientPermissionDenied(reason)

        self.logger.error(
            "provider_key: %s, path: %s\n\tBad result from das service. Message: %s",
            self.provider_key,
            path,
            response.text,
        )
        raise DasClientException("Failed to post to DAS web service.")

    def post_event_photo(self, event_id, image):

        raise ValueError("post_event_photo is no longer valid.")
        photos_path = "activity/event/" + str(event_id) + "/photos/"

        with open(image, "rb") as image_file:
            files = {"image": image_file}
            return self._post_form(photos_path, files=files)

    def post_camera_trap_report(self, camera_trap_payload, file=None):

        camera_trap_report_path = (
            f"sensors/camera-trap/" + self.provider_key + "/status/"
        )

        if file:
            files = {"filecontent.file": file}
            return self._post_form(
                camera_trap_report_path, body=camera_trap_payload, files=files
            )
        else:
            file_path = camera_trap_payload.get("file")

            with open(file_path, "rb") as f:
                files = {"filecontent.file": f}
            return self._post_form(
                camera_trap_report_path, body=camera_trap_payload, files=files
            )

    def delete_event_file(self, event_id, file_id):
        self._delete(f"activity/event/{event_id}/file/{file_id}")

    def delete_event_note(self, event_id, note_id):

        path = f"activity/event/{event_id}/note/{note_id}"
        self._delete(path)

    def post_event_file(self, event_id, filepath=None, comment=""):

        documents_path = "activity/event/" + str(event_id) + "/files/"

        with open(filepath, "rb") as f:
            files = {"filecontent.file": f}
            return self._post_form(
                documents_path, body={"comment": comment}, files=files
            )

    def post_event_note(self, event_id, notes):

        created = []

        if not isinstance(notes, list):
            notes = [
                notes,
            ]

        for note in notes:
            notesRequest = {"event": event_id, "text": note}

            result = self._post("activity/event/" + event_id + "/notes", notesRequest)
            created.append(result)

        return created

    def get_me(self):
        """
        Get details for the 'me', the current DAS user.
        :return:
        """
        return self._get("user/me")

    def post_subject(self, subject):
        """
        Post a subject payload to create a new subject.
        :param subject:
        :return:
        """
        self.logger.debug(f"Posting subject {subject.get('name')}")
        return self._post("subjects", payload=subject)

    def post_source(self, source):
        """
        Post a source payload to create a new source.
        :param source:
        :return:
        """
        self.logger.debug(
            "Posting source for manufacturer_id: %s", source.get("manufacturer_id")
        )
        return self._post("sources", payload=source)

    def _clean_observation(self, observation):
        if hasattr(observation["recorded_at"], "isoformat"):
            observation["recorded_at"] = observation["recorded_at"].isoformat()
        return observation

    def _clean_event(self, event):
        return event

    def post_radio_observation(self, observation):
        # Clean-up data before posting
        observation["recorded_at"] = observation["recorded_at"].isoformat()
        self.logger.debug("Posting observation: %s", observation)
        result = self._post(
            "sensors/dasradioagent/{}/status".format(self.provider_key),
            payload=observation,
        )
        self.logger.debug("Result of post is: %s", result)
        return result

    def post_radio_heartbeat(self, data):
        self.logger.debug("Posting heartbeat: %s", data)
        result = self._post(
            "sensors/dasradioagent/{}/status".format(self.provider_key), payload=data
        )
        self.logger.debug("Result of heartbeat post is: %s", result)

    def post_observation(self, observation):
        """
        Post a new observation, or a list of observations.
        """
        if isinstance(observation, (list, set)):
            payload = [self._clean_observation(o) for o in observation]
        else:
            payload = self._clean_observation(observation)

        self.logger.debug("Posting observation: %s", payload)
        return self._post("observations", payload=payload)

    def post_sensor_observation(self, observation, sensor_type="generic"):
        """
        Post a new observation, or a list of observations.
        """
        if isinstance(observation, (list, set)):
            payload = [self._clean_observation(o) for o in observation]
        else:
            payload = self._clean_observation(observation)

        self.logger.debug("Posting observation: %s", observation)
        result = self._post(
            "sensors/{}/{}/status".format(sensor_type, self.provider_key),
            payload=observation,
        )
        self.logger.debug("Result of post is: %s", result)
        return result

    def post_patrol(self, data):
        payload = self._clean_event(data)
        self.logger.debug("Posting patrol: %s", payload)
        result = self._post("activity/patrols", payload=payload)
        self.logger.debug("Result of patrol post is: %s", result)
        return result

    def patch_event_type(self, event_type):
        self.logger.debug("Patching event type: %s", event_type)
        result = self._patch(
            f"activity/events/eventtypes/{event_type['id']}", payload=event_type
        )
        self.logger.debug("Result of event type patch is: %s", result)
        return result

    def post_event_type(self, event_type):
        self.logger.debug("Posting event type: %s", event_type)
        result = self._post("activity/events/eventtypes/", payload=event_type)
        self.logger.debug("Result of event type post is: %s", result)
        return result

    def post_report(self, data):
        payload = self._clean_event(data)
        self.logger.debug("Posting report: %s", payload)
        result = self._post("activity/events", payload=payload)
        self.logger.debug("Result of report post is: %s", result)
        return result

    def post_event_category(self, data):
        self.logger.debug("Posting event category: %s", data)
        result = self._post("activity/events/categories", payload=data)
        self.logger.debug("Result of report category post is: %s", result)
        return result

    def patch_event_category(self, data):
        self.logger.debug("Patching event category: %s", data)
        result = self._patch(f'activity/events/categories/{data["id"]}', payload=data)
        self.logger.debug("Result of report category patch is: %s", result)
        return result

    def post_event(self, event):
        """
        Post a new Event.
        """
        return self.post_report(event)

    def add_events_to_patrol_segment(self, events, patrol_segment):
        for event in events:
            payload = {"id": event["id"], "patrol_segments": [patrol_segment["id"]]}

            result = self._patch(f"activity/event/{event['id']}", payload=payload)

    def patch_event(self, event_id, payload):
        self.logger.debug("Patching event: %s", payload)
        result = self._patch("activity/event/" + event_id, payload=payload)
        self.logger.debug("Result of event patch is: %s", result)
        return result

    def get_file(self, url):
        return self._get(url, stream=True, return_response=True)

    def get_event_type(self, event_type_name):
        return self._get(f"activity/events/schema/eventtype/{event_type_name}")

    def get_event_categories(self, include_inactive=False):
        return self._get(
            f"activity/events/categories", params={"include_inactive": include_inactive}
        )

    def get_messages(self):

        results = self._get(path="messages")

        while True:
            if results and results.get("results"):
                for r in results["results"]:
                    yield r

            if results and results["next"]:
                url, params = split_link(results["next"])
                p["page"] = params["page"]
                results = self._get(path="messages")
            else:
                break

    def get_event_types(self, **params):
        return self._get("activity/events/eventtypes", params=params)

    def get_event_schema(self, event_type):
        return self._get(f"activity/events/schema/eventtype/{event_type}")

    def _get_objects_count(self, params):
        params = params.copy()
        params["page"] = 1
        params["page_size"] = 1
        events = self._get(params["object"], params=params)
        if events and events.get("count"):
            return events["count"]
        return 0

    def get_objects(self, **kwargs):
        params = dict((k, v) for k, v in kwargs.items() if k not in ("page"))
        if not params.get("object"):
            raise ValueError("Must specify object URL")

        result = self._get(params["object"], params=params)
        for o in result:
            yield o

        self.logger.debug(f"Getting {params['object']}: ", params)

        count = 0
        results = self._get(params["object"], params=params)

        while True:
            if not results:
                break

            if "results" in results:
                for result in results["results"]:
                    yield result
                    count += 1
                    if ("max_results" in params) and (count >= params["max_results"]):
                        return
                next = results.get("next")
                if next and ("page" not in params):
                    url = re.sub(f".*{params['object']}?", params["object"], next)
                    self.logger.debug("Getting more events: " + url)
                    results = self._get(url)

                else:
                    break
            else:
                for o in result:
                    yield o
                break

    def get_objects_multithreaded(self, **kwargs):
        threads = kwargs.get("threads", 5)
        params = dict((k, v) for k, v in kwargs.items() if k not in ("page"))
        if not params.get("object"):
            raise ValueError("Must specify object URL")

        if not params.get("page_size"):
            params["page_size"] = 100

        count = self._get_objects_count(params)

        self.logger.debug(
            f"Loading {count} {params['object']} from ER with page size {params['page_size']} and {threads} threads"
        )
        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            futures = []
            for page in range(1, math.ceil(count / params["page_size"]) + 1):
                temp_params = params.copy()
                temp_params["page"] = page
                futures.append(
                    executor.submit(self._get, params["object"], params=temp_params)
                )
            for future in concurrent.futures.as_completed(futures):
                try:
                    result = future.result()
                    for e in result["results"]:
                        yield e
                except Exception as e:
                    logging.error(f"Error occurred loading events: {e}")
                    raise e

    def get_events(self, **kwargs):
        params = dict(
            (k, v)
            for k, v in kwargs.items()
            if k
            in (
                "state",
                "page_size",
                "page",
                "event_type",
                "filter",
                "include_notes",
                "include_related_events",
                "include_files",
                "include_details",
                "updated_since",
                "include_updates",
                "max_results",
                "oldest_update_date",
                "event_ids",
            )
        )

        self.logger.debug("Getting events: ", params)
        events = self._get("activity/events", params=params)

        count = 0
        while True:
            if events and events.get("results"):
                for result in events["results"]:
                    yield result
                    count += 1
                    if ("max_results" in params) and (count >= params["max_results"]):
                        return
            if events["next"] and ("page" not in params):
                url = events["next"]
                url = re.sub(".*activity/events?", "activity/events", events["next"])
                self.logger.debug("Getting more events: " + url)
                events = self._get(url)
            else:
                break

    def get_patrols(self, **kwargs):
        params = dict(
            (k, v)
            for k, v in kwargs.items()
            if k in ("state", "page_size", "page", "event_type", "filter")
        )
        self.logger.debug("Getting patrols: ", params)
        patrols = self._get("activity/patrols", params=params)

        while True:
            if patrols and patrols.get("results"):
                for result in patrols["results"]:
                    yield result
            if patrols["next"]:
                url = patrols["next"]
                url = re.sub(".*activity/patrols?", "activity/patrols", patrols["next"])
                self.logger.debug("Getting more patrols: " + url)
                patrols = self._get(url)
            else:
                break

    def get_events_export(self, filter=None):
        params = None
        if filter:
            params = {"filter": filter}

        response = self._get(
            "activity/events/export/", params=params, return_response=True
        )
        return response

    def pulse(self, message=None):
        """
        Convenience method for getting status of the DAS api.
        :param message:
        :return:
        """
        return self._get("status")

    def get_subject_sources(self, subject_id):
        return self._get(path=f"subject/{subject_id}/sources")

    def get_subjectsources(self, subject_id):
        return self._get(path=f"subject/{subject_id}/subjectsources")

    def get_source_provider(self, provider_key):
        results = self._get("sourceproviders")

        while True:
            if results and results.get("results"):
                for r in results["results"]:
                    if r.get("provider_key") == provider_key:
                        return r

            if results and results["next"]:
                url, params = split_link(results["next"])
                results = self._get(path="sourceproviders", params=params)
            else:
                break

        return None

    def get_subject_tracks(self, subject_id="", start=None, end=None):
        """
        Get the latest tracks for the Subject having the given subject_id.
        """
        p = {}
        if start is not None and isinstance(start, datetime):
            p["since"] = start.isoformat()
        if end is not None and isinstance(end, datetime):
            p["until"] = end.isoformat()

        return self._get(path="subject/{0}/tracks".format(subject_id), params=p)

    def get_subject_trackingdata(
        self,
        subject_id=None,
        subject_chronofile=None,
        include_inactive=True,
        start=None,
        end=None,
        out_format="json",
        filter_flag=0,
        current_status=False,
    ):
        p = {}
        if start is not None and isinstance(start, datetime):
            p["after_date"] = start.isoformat()
        if end is not None and isinstance(end, datetime):
            p["before_date"] = end.isoformat()
        if subject_id:
            p["subject_id"] = subject_id
        elif subject_chronofile:
            p["subject_chronofile"] = subject_chronofile
        else:
            raise ValueError("specify subject_id or subject_chronofile")
        p["include_inactive"] = include_inactive
        p["format"] = out_format  # should be 'json' or 'csv'
        p["filter"] = "null" if filter_flag is None else filter_flag
        p["current_status"] = current_status
        return self._get(path="trackingdata/export", params=p)

    def get_subject_trackingmetadata(self, include_inactive=True, out_format="json"):
        p = {}
        p["include_inactive"] = include_inactive
        p["format"] = out_format  # should be 'json' or 'csv'
        return self._get(path="trackingmetadata/export", params=p)

    def get_subject_observations(
        self,
        subject_id,
        start=None,
        end=None,
        filter_flag=0,
        include_details=True,
        page_size=10000,
    ):
        return self.get_observations(
            subject_id=subject_id,
            start=start,
            end=end,
            filter_flag=filter_flag,
            include_details=include_details,
            page_size=page_size,
        )

    def get_source_observations(
        self,
        source_id,
        start=None,
        end=None,
        filter_flag=0,
        include_details=True,
        page_size=10000,
    ):
        return self.get_observations(
            source_id=source_id,
            start=start,
            end=end,
            filter_flag=filter_flag,
            include_details=include_details,
            page_size=page_size,
        )

    def get_observations(
        self,
        subject_id=None,
        source_id=None,
        start=None,
        end=None,
        filter_flag=0,
        include_details=True,
        page_size=10000,
    ):
        p = {}
        if start is not None and isinstance(start, datetime):
            p["since"] = start.isoformat()
        if end is not None and isinstance(end, datetime):
            p["until"] = end.isoformat()
        if subject_id:
            p["subject_id"] = subject_id
        elif source_id:
            p["source_id"] = source_id

        p["filter"] = "null" if filter_flag is None else filter_flag
        p["include_details"] = include_details
        p["page_size"] = page_size  # current limit

        results = self._get(path="observations", params=p)

        while True:
            if results and results.get("results"):
                for r in results["results"]:
                    yield r

            if results and results["next"]:
                url, params = split_link(results["next"])
                p["page"] = params["page"]
                results = self._get(path="observations", params=p)
            else:
                break

    def get_subjects(self, subject_group_id=None, **kwargs):
        """
        Get the list of subjects to whom the user has access.
        :return:
        """
        params = dict(
            (k, v)
            for k, v in kwargs.items()
            if k in ("subject_group", "include_inactive")
        )

        return self._get("subjects", params=params)

    def get_subject(self, subject_id=""):
        """
        get the subject given the subject id
        :param subject_id: the UUID for the subject
        :return:
        """
        return self._get(path="subject/{0}".format(subject_id))

    def get_source_by_id(self, id):
        """
        get the source by id
        :param id: source id
        :return:
        """
        return self._get(path="source/{0}".format(id))

    def get_source_by_manufacturer_id(self, id):
        """
        get the source by manufacturer id or collar id
        :param id: the manufacturer id
        :return:
        """
        return self._get(path="source/{0}".format(id))

    def get_subjectgroups(
        self,
        include_inactive=False,
        include_hidden=True,
        isvisible=True,
        flat=True,
        group_name=None,
    ):
        """Get the list of visible subjectgroups including members.
         By default don't include inactive subjects
         to get all subject groups whether visible or not, call with include_hidden=True

        Args:
            include_inactive (bool, optional): set to True to include inactive subjects. Defaults to False.
            include_hidden (bool, optional): include subject groups that are not visible (isvisible flag is false). Defaults to True.
            isvisible (bool, optional): either include all visible groups, or only include not visible groups. Defaults to True.
            flat (bool, optional): unnest parent/child subjectgroups returning a flat list of subjectgroups
            group_name (string, optional): filter the subjectgroups to this name

        Returns:
            [type]: [description]
        """
        p = dict()
        p["include_inactive"] = include_inactive
        p["include_hidden"] = include_hidden
        p["isvisible"] = isvisible
        p["flat"] = flat
        p["group_name"] = group_name

        return self._get("subjectgroups", params=p)

    def get_sources(self, page_size=100):
        """Return all sources"""
        params = dict(page_size=page_size)
        sources = "sources"
        results = self._get(path=sources, params=params)

        while True:
            if results and results.get("results"):
                for r in results["results"]:
                    yield r

            if results and results["next"]:
                _, qparam = split_link(results["next"])
                params["page"] = qparam["page"]
                results = self._get(path=sources, params=params)
            else:
                break

    def get_users(self):
        return self._get("users")


class DasClientException(Exception):
    pass


class DasClientPermissionDenied(DasClientException):
    pass


class DasClientServiceUnavailable(DasClientException):
    pass


class DasClientNotFound(DasClientException):
    pass



from ecoscope.analysis import UD, astronomy, seasons
from ecoscope.analysis.ecograph import Ecograph, get_feature_gdf
from ecoscope.analysis.percentile import get_percentile_area
from ecoscope.analysis.speed import SpeedDataFrame

__all__ = [
    "Ecograph",
    "SpeedDataFrame",
    "UD",
    "astronomy",
    "get_feature_gdf",
    "get_percentile_area",
    "seasons",
]



import datetime
import typing
from dataclasses import dataclass, field

import geopandas
import shapely
import shapely.geometry


@dataclass
class RelocsCoordinateFilter:
    """Filter parameters for filtering get_fixes based on X/Y coordinate ranges or specific coordinate values"""

    min_x: float = -180.0
    max_x: float = 180.0
    min_y: float = -90.0
    max_y: float = 90.0
    filter_point_coords: typing.Union[typing.List[typing.List[float]], geopandas.GeoSeries] = field(
        default_factory=[[0.0, 0.0]]
    )

    def __post_init__(self):
        if isinstance(self.filter_point_coords, list):
            self.filter_point_coords = geopandas.GeoSeries(
                shapely.geometry.Point(coord) for coord in self.filter_point_coords
            )


@dataclass
class RelocsDateRangeFilter:
    """
    Filter parameters for filtering based on a datetime range
    """

    start: datetime.datetime
    end: datetime.datetime


@dataclass
class RelocsSpeedFilter:
    """
    Filter parameters for filtering based on the speed needed to move from one fix to the next
    """

    max_speed_kmhr: float = float("inf")
    temporal_order: str = "ASC"


@dataclass
class RelocsDistFilter:
    """
    Filter based on the distance between consecutive fixes. Fixes are filtered to the range [min_dist_km, max_dist_km].
    """

    min_dist_km: float = 0.0
    max_dist_km: float = float("inf")
    temporal_order: str = "ASC"


@dataclass
class TrajSegFilter:
    """
    Class filtering a set of trajectory segment segments
    """

    min_length_meters: float = 0.0
    max_length_meters: float = float("inf")
    min_time_secs: float = 0.0
    max_time_secs: float = float("inf")
    min_speed_kmhr: float = 0.0
    max_speed_kmhr: float = float("inf")



import os
import typing
from dataclasses import dataclass, field

import geopandas as gpd
import numpy as np
import rasterio
import rasterio.features
from shapely.geometry import shape
from shapely.geometry.multipolygon import MultiPolygon


@dataclass
class PercentileAreaProfile:
    input_raster: typing.Union[str, bytes, os.PathLike]
    percentile_levels: typing.List = field(default_factory=[50.0])
    subject_id: str = ""


class PercentileArea:
    @staticmethod
    def _multipolygon(shapes, percentile):
        return MultiPolygon([shape(geom) for geom, value in shapes if np.isclose(value, percentile)])

    @classmethod
    def calculate_percentile_area(cls, profile: PercentileAreaProfile):
        """
        Parameters
        ----------
        profile:  PercentileAreaProfile
            dataclass object with information for percentile-area calculation

        Returns
        -------
        GeodataFrame
        """

        assert type(profile) is PercentileAreaProfile
        shapes = []

        # open raster
        with rasterio.open(profile.input_raster) as src:
            crs = src.crs.to_wkt()

            for percentile in profile.percentile_levels:
                data_array = src.read(1).astype(np.float32)

                # Mask no-data values
                data_array[data_array == src.nodata] = np.nan

                # calculate percentile value
                # percentile_val = np.percentile(data_array[~np.isnan(data_array)], 100.0 - percentile)
                values = np.sort(data_array[~np.isnan(data_array)]).flatten()
                csum = np.cumsum(values)
                percentile_val = values[np.argmin(np.abs(csum[-1] * (1 - percentile / 100) - csum))]

                # TODO: make a more explicit comparison for less than and greater than

                # Set any vals less than the cutoff to be nan
                data_array[data_array < percentile_val] = np.nan

                # Mask any vals that are less than the cutoff percentile
                data_array[data_array >= percentile_val] = percentile

                shapes.extend(rasterio.features.shapes(data_array, transform=src.transform))

        return gpd.GeoDataFrame(
            [
                [profile.subject_id, percentile, cls._multipolygon(shapes, percentile)]
                for percentile in sorted(profile.percentile_levels, reverse=True)
            ],
            columns=["subject_id", "percentile", "geometry"],
            crs=crs,
        )


def get_percentile_area(
    percentile_levels: typing.List,
    raster_path: typing.Union[str, bytes, os.PathLike],
    subject_id: str = "",
):
    """
    Parameters
    ----------
    percentile_levels: Typing.List[Int]
        list of k-th percentile scores.
    raster_path: str or os.PathLike
        file path to where the raster is stored.
    subject_id: str
        unique identifier for the subject

    Returns
    -------
    GeoDataFrame

    """
    percentile_profile = PercentileAreaProfile(
        percentile_levels=percentile_levels,
        input_raster=raster_path,
        subject_id=subject_id,
    )
    return PercentileArea.calculate_percentile_area(profile=percentile_profile)



from ecoscope import analysis, base, contrib, io, mapping, plotting

ASCII = """\
 _____
|   __|___ ___ ___ ___ ___ ___ ___
|   __|  _| . |_ -|  _| . | . | -_|
|_____|___|___|___|___|___|  _|___|
                          |_|
"""

__initialized = False


def init(silent=False, selenium=False, force=False):
    """
    Initializes the environment with ecoscope-specific customizations.

    Parameters
    ----------
    silent : bool, optional
        Removes console output
    selenium : bool, optional
        Installs selenium webdriver in a colab environment
    force : bool, optional
        Ignores `__initialized`

    """

    global __initialized
    if __initialized and not force:
        if not silent:
            print("Ecoscope already initialized.")
        return

    import pandas as pd

    pd.options.plotting.backend = "plotly"

    from tqdm.auto import tqdm

    tqdm.pandas()

    import warnings

    from shapely.errors import ShapelyDeprecationWarning

    warnings.filterwarnings(action="ignore", category=ShapelyDeprecationWarning)
    warnings.filterwarnings(action="ignore", category=FutureWarning)
    warnings.filterwarnings("ignore", message=".*initial implementation of Parquet.*")

    import geopandas as gpd

    def explore(data, *args, **kwargs):
        """
        Monkey-patched `geopandas.explore._explore` to use EcoMap instead.
        """
        initialized = "m" in kwargs
        if not initialized:
            kwargs["m"] = mapping.EcoMap()

        if isinstance(kwargs["m"], mapping.EcoMap):
            m = kwargs.pop("m")
            m.add_gdf(data, *args, **kwargs)
            if not initialized:
                m.zoom_to_bounds(data.geometry.to_crs(4326).total_bounds)
            return m
        else:
            return gpd.explore._explore(data, *args, **kwargs)

    gpd.GeoDataFrame.explore = explore
    gpd.GeoSeries.explore = explore

    import plotly.io as pio

    pio.templates.default = "seaborn"

    import sys

    if "google.colab" in sys.modules and selenium:
        import subprocess

        subprocess.run(["apt", "update"])
        subprocess.run(["apt", "install", "chromium-chromedriver"])

    __initialized = True
    if not silent:
        print(ASCII)


__all__ = [
    "analysis",
    "base",
    "contrib",
    "init",
    "io",
    "mapping",
    "plotting",
]



import base64
import os
import tempfile
import time
import typing
import urllib
import warnings

import ee
import folium
import geopandas as gpd
import matplotlib as mpl
import numpy as np
import pandas as pd
import rasterio
import selenium.webdriver
from branca.colormap import StepColormap
from branca.element import MacroElement, Template

import ecoscope
from ecoscope.contrib.foliumap import Map

warnings.filterwarnings("ignore", "GeoSeries.isna", UserWarning)


class EcoMapMixin:
    def add_speedmap(
        self,
        trajectory: gpd.GeoDataFrame,
        classification_method: str = "equal_interval",
        num_classes: int = 6,
        speed_colors: typing.List = None,
        bins: typing.List = None,
        legend: bool = True,
    ):

        speed_df = ecoscope.analysis.SpeedDataFrame.from_trajectory(
            trajectory=trajectory,
            classification_method=classification_method,
            num_classes=num_classes,
            speed_colors=speed_colors,
            bins=bins,
        )
        self.add_gdf(speed_df, color=speed_df["speed_colour"])

        if legend:
            self.add_legend(legend_dict=dict(zip(speed_df.label, speed_df.speed_colour)))

        return speed_df


class EcoMap(EcoMapMixin, Map):
    def __init__(self, *args, static=False, print_control=True, **kwargs):
        kwargs["attr"] = kwargs.get("attr", " ")
        kwargs["canvas"] = kwargs.get("canvas", True)
        kwargs["control_scale"] = kwargs.get("control_scale", True)
        kwargs["height"] = kwargs.get("height", 600)
        kwargs["width"] = kwargs.get("width", 800)

        if static:
            print_control = False
            kwargs["draw_control"] = kwargs.get("draw_control", False)
            kwargs["fullscreen_control"] = kwargs.get("fullscreen_control", False)
            kwargs["layers_control"] = kwargs.get("layers_control", False)
            kwargs["measure_control"] = kwargs.get("measure_control", False)
            kwargs["zoom_control"] = kwargs.get("zoom_control", False)
            kwargs["search_control"] = kwargs.get("search_control", False)

        self.px_height = kwargs["height"]
        self.px_width = kwargs["width"]

        super().__init__(*args, **kwargs)

        if print_control:
            self.add_print_control()

    def add_gdf(self, data, *args, simplify_tolerance=None, **kwargs):
        """
        Wrapper for `geopandas.explore._explore`.
        """

        data = data.copy()
        data = data.to_crs(4326)
        data = data.loc[(~data.geometry.isna()) & (~data.geometry.is_empty)]

        if simplify_tolerance is not None:
            data = data.simplify(simplify_tolerance)

        if isinstance(data, gpd.GeoDataFrame):
            for col in data:
                if pd.api.types.is_datetime64_any_dtype(data[col]):
                    data[col] = data[col].astype("string")

        kwargs["m"] = self
        kwargs["tooltip"] = kwargs.get("tooltip", False)
        gpd.explore._explore(data, *args, **kwargs)

    def add_legend(self, *args, **kwargs):
        """
        Patched method for allowing legend hex colors to start with a "#".
        """
        legend_dict = kwargs.get("legend_dict")
        if legend_dict is not None:
            kwargs["legend_dict"] = {
                k: v[1:] if isinstance(v, str) and v.startswith("#") else v for k, v in legend_dict.items()
            }

        return super().add_legend(*args, **kwargs)

    def add_north_arrow(self, position="topright", scale=1.0):
        """
        Parameters
        ----------
        position : str
            Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.
        scale : float
            Scale dimensions of north arrow.

        """

        self.add_child(
            ControlElement(
                f"""\
            <svg width="{100*scale}px" height="{100*scale}px" version="1.1" viewBox="0 0 773 798" xmlns="http://www.w3.org/2000/svg">
<path transform="translate(0 798) scale(1 -1)" d="m674 403-161 48q-17 48-66 70l-46 166-46-167q-22-9-38-25t-29-45l-159-47 159-49q15-44 67-68l46-164 48 164q39 17 64 69zm-163 0q0-49-32-81-33-34-78-34-46 0-77 34-31 31-31 81 0 46 31 80t77 34q45 0 78-34 32-34 32-80zm-12 1q-5 7-7.5 17.5t-4 21.5-4.5 21-9 16q-7 6-17 9.5t-20.5 6-20 6-15.5 9.5v-107h98zm-98-108v108h-99l3-3 23-75q6-6 16.5-9.5t21-5.5 20-5.5 15.5-9.5zm-280 152h-26v-2q5 0 6-1 3-3 3-6 0-2-0.5-4t-1.5-7l-18-48-16 47q-3 9-3 12 0 7 7 7h2v2h-34v-2q2 0 3-1l3-3q2 0 2-2 2-1 4-5l5-15-12-42-17 50q-3 9-3 11 0 7 6 7h2v2h-33v-2q8 0 10-6 1-2 3-9l27-74h5l15 53 19-53h2l27 71q2 10 3 11 5 7 10 7v2zm325 350h-29v-3q7 0 10-4 1-1 1-11v-35l-42 53h-32v-3q7-2 12-6l2-3v-62q0-13-12-13v-2h29v2h-2q-4 0-7 2.5t-3 10.5v55l58-72h3v73q0 9 1 10.5t8 3.5l3 1v3zm207-395h-130q0 16-6 42zm-212-119-40-141v135q9 0 19 1t21 5zm-154 78-137 41h130q0-10 2-19.5t5-21.5zm114 168q-25 0-39-8l39 142v-134zm372-148h-3q-3-4-5-7.5t-4-5.5q-5-5-17-5h-19q-3 0-3 5v35h20q8 0 10-6 1-1 1-3 0-3 1-4h3v30h-3q-2-9-4-11t-8-2h-20v35h24q7 0 8-1 4-1 9-14h3l-1 20h-69v-2h3q7 0 8-4 2-2 2-9v-58q0-11-4-12-1-1-6-1h-3v-3h68zm-340-358q0 9-5.5 14.5t-20.5 14.5q-9 5-13 9l-5 5q-3 10-3 7 0 14 14 14 18 0 24-26h2v31h-2q-2-6-5-6-4 0-5 1-8 5-15 5-11 0-17.5-7t-6.5-17q0-13 9-19 6-4 16.5-10.5t12.5-8.5q8-7 8-13 0-14-18-14-13 0-18 5.5t-7 20.5h-2v-30h2q0 5 3 5l16-5h8q12 0 20 7t8 17z"/>
            </svg>""",  # noqa
                position=position,
            )
        )

    def add_title(self, title, font_size="32px", **kwargs):
        """
        Parameters
        ----------
        title : str
            Text of title.
        font_size : str
            CSS font size that includes units.
        kwargs
            Additional style kwargs. Underscores in keys will be replaced with dashes

        """
        title_html = f"""\
        <div style="position: absolute; left: 50%;">
            <p style="position: relative;
                    left: -50%;
                    border: 1px solid #000;
                    border-radius: 5px;
                    background-color: #FFFFFF99;
                    padding: 3px;
                    font-size: { font_size };
                    { "".join([f"{k.replace('_', '-')}: {v};" for k, v in kwargs.items() ]) }
                    ">
            { title }
            </p>
        </div>
        """
        self.add_child(FloatElement(title_html, top=0, left=0, right=0))

    def _repr_html_(self, **kwargs):
        return (
            super()
            ._repr_html_(**kwargs)
            .replace(
                urllib.parse.quote("crs: L.CRS."),
                urllib.parse.quote("attributionControl: false, crs: L.CRS."),
            )
        )

    def to_html(self, outfile, **kwargs):
        """
        Parameters
        ----------
        outfile : str, Pathlike
            Output destination

        """
        with open(outfile, "w") as file:
            file.write(self._repr_html_(**kwargs))

    def to_png(self, outfile, sleep_time=10, **kwargs):
        """
        Parameters
        ----------
        outfile : str, Pathlike
            Output destination
        sleep_time : int, optional
            Additional seconds to wait before taking screenshot. Should be increased if map tiles in the output haven't
            fully loaded but can also be decreased in most cases.

        """

        with tempfile.NamedTemporaryFile(mode="w", suffix=".html") as tmp:
            super().to_html(tmp.name, **kwargs)
            chrome_options = selenium.webdriver.chrome.options.Options()
            chrome_options.add_argument("--headless")
            chrome_options.add_argument("--no-sandbox")
            chrome_options.add_argument("--disable-dev-shm-usage")
            driver = selenium.webdriver.Chrome(options=chrome_options)
            if self.px_width and self.px_height:
                driver.set_window_size(width=self.px_width, height=self.px_height)
            driver.get(f"file://{os.path.abspath(tmp.name)}")
            time.sleep(sleep_time)
            driver.save_screenshot(outfile)

    def add_ee_layer(self, ee_object, visualization_params, name) -> None:
        """
        Method for displaying Earth Engine image tiles.

        Parameters
        ----------
        ee_object
        visualization_params
        name

        Returns
        -------
        None

        """

        try:
            if isinstance(ee_object, ee.image.Image):
                map_id_dict = ee.Image(ee_object).getMapId(visualization_params)
                folium.raster_layers.TileLayer(
                    tiles=map_id_dict["tile_fetcher"].url_format,
                    attr="Google Earth Engine",
                    name=name,
                    overlay=True,
                    control=True,
                ).add_to(self)

            elif isinstance(ee_object, ee.imagecollection.ImageCollection):
                ee_object_new = ee_object.mosaic()
                map_id_dict = ee.Image(ee_object_new).getMapId(visualization_params)
                folium.raster_layers.TileLayer(
                    tiles=map_id_dict["tile_fetcher"].url_format,
                    attr="Google Earth Engine",
                    name=name,
                    overlay=True,
                    control=True,
                ).add_to(self)

            elif isinstance(ee_object, ee.geometry.Geometry):
                folium.GeoJson(data=ee_object.getInfo(), name=name, overlay=True, control=True).add_to(self)

            elif isinstance(ee_object, ee.featurecollection.FeatureCollection):
                ee_object_new = ee.Image().paint(ee_object, 0, 2)
                map_id_dict = ee.Image(ee_object_new).getMapId(visualization_params)
                folium.raster_layers.TileLayer(
                    tiles=map_id_dict["tile_fetcher"].url_format,
                    attr="Google Earth Engine",
                    name=name,
                    overlay=True,
                    control=True,
                ).add_to(self)

        except Exception as exc:
            print(f"{exc}. Could not display {name}")

    def zoom_to_bounds(self, bounds):
        """Zooms to a bounding box in the form of [minx, miny, maxx, maxy].

        Parameters
        ----------
        bounds : [x1, y1, x2, y2]
            Map extent in WGS 84
        """

        assert -180 < bounds[0] <= bounds[2] < 180
        assert -90 < bounds[1] <= bounds[3] < 90

        self.fit_bounds([[bounds[1], bounds[0]], [bounds[3], bounds[2]]])

    def zoom_to_gdf(self, gs):
        """
        Parameters
        ----------
        gs : gpd.GeoSeries or gpd.GeoDataFrame
            Geometry to adjust map bounds to. CRS will be converted to WGS 84.

        """

        self.zoom_to_bounds(gs.geometry.to_crs(4326).total_bounds)

    def add_local_geotiff(self, path, zoom=False, cmap=None, colorbar=True):
        """
        Displays a local GeoTIFF.

        Parameters
        ----------
        path : str, Pathlike
            Path to local GeoTIFF
        zoom : bool
            Zoom to displayed GeoTIFF
        cmap : `matplotlib.colors.Colormap` or str or None
            Matplotlib colormap to apply to raster
        colorbar : bool
            Whether to add colorbar for provided `cmap`. Does nothing if `cmap` is not provided.
        """

        with rasterio.open(path) as src:
            transform, width, height = rasterio.warp.calculate_default_transform(
                src.crs, "EPSG:4326", src.width, src.height, *src.bounds
            )
            kwargs = src.meta.copy()
            kwargs.update({"crs": "EPSG:4326", "transform": transform, "width": width, "height": height})

            if cmap is None:
                im = [rasterio.band(src, i + 1) for i in range(src.count)]
            else:
                cmap = mpl.cm.get_cmap(cmap)
                kwargs["count"] = 4
                im = rasterio.band(src, 1)[0].read()[0]
                im_min, im_max = np.nanmin(im), np.nanmax(im)
                im = np.rollaxis(cmap((im - im_min) / (im_max - im_min), bytes=True), -1)
                if colorbar:
                    if isinstance(im_min, np.integer) and im_max - im_min < 256:
                        self.add_child(
                            StepColormap(
                                [mpl.colors.rgb2hex(color) for color in cmap(np.linspace(0, 1, 1 + im_max - im_min))],
                                index=np.arange(1 + im_max - im_min),
                                vmin=im_min,
                                vmax=im_max + 1,
                            )
                        )
                    else:
                        self.add_child(
                            StepColormap(
                                [mpl.colors.rgb2hex(color) for color in cmap(np.linspace(0, 1, 256))],
                                vmin=im_min,
                                vmax=im_max,
                            )
                        )

            with rasterio.io.MemoryFile() as memfile:
                with memfile.open(**kwargs) as dst:
                    for i in range(kwargs["count"]):
                        rasterio.warp.reproject(
                            source=im[i],
                            destination=rasterio.band(dst, i + 1),
                            src_transform=src.transform,
                            src_crs=src.crs,
                            dst_transform=transform,
                            dst_crs="EPSG:4326",
                            resampling=rasterio.warp.Resampling.nearest,
                        )

                url = "data:image/tiff;base64," + base64.b64encode(memfile.read()).decode("utf-8")

        self.add_child(GeoTIFFElement(url, zoom))

    def add_print_control(self):
        self.add_child(PrintControl())


class ControlElement(MacroElement):
    """
    Class to wrap arbitrary HTML as Leaflet Control.

    Parameters
    ----------
    html : str
        HTML to render an element from.
    position : str
        Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.

    """

    _template = Template(
        """
        {% macro script(this, kwargs) %}
        var {{ this.get_name() }} = L.Control.extend({
            onAdd: function(map) {
                var template = document.createElement('template');
                template.innerHTML = `{{ this.html }}`.trim();
                return template.content.firstChild;
            }
        });
        (new {{ this.get_name() }}({{ this.options|tojson }})).addTo({{this._parent.get_name()}});

        {% endmacro %}
    """
    )

    def __init__(self, html, position="bottomright"):
        super().__init__()
        self.html = html
        self.options = folium.utilities.parse_options(
            position=position,
        )


class FloatElement(MacroElement):
    """
    Class to wrap arbitrary HTML as a floating Element.

    Parameters
    ----------
    html : str
        HTML to render an element from.
    left, right, top, bottom : str
        Distance between edge of map and nearest edge of element. Two should be provided. Style can also be specified
        in html.

    """

    _template = Template(
        """
        {% macro header(this,kwargs) %}
        <style>
            #{{ this.get_name() }} {
                position: absolute;
                z-index: 9999;
                left: {{ this.left }};
                right: {{ this.right }};
                top: {{ this.top }};
                bottom: {{ this.bottom }};
            }
        </style>
        {% endmacro %}

        {% macro html(this,kwargs) %}
        <div id="{{ this.get_name() }}">
             {{ this.html }}
        </div>
        {% endmacro %}
    """
    )

    def __init__(self, html, left="", right="", top="", bottom=""):
        super().__init__()
        self.html = html
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom


class GeoTIFFElement(MacroElement):
    """
    Class to display a GeoTIFF.

    Parameters
    ----------
    url : str
        URL of GeoTIFF
    zoom : bool
        Zoom to displayed GeoTIFF
    """

    _template = Template(
        """
    {% macro html(this, kwargs) %}
        <script src="https://unpkg.com/georaster-layer-for-leaflet@3.8.0/dist/georaster-layer-for-leaflet.min.js"></script>
        <script src="https://unpkg.com/georaster@1.5.6/dist/georaster.browser.bundle.min.js"></script>
    {% endmacro %}

    {% macro script(this, kwargs) %}
        fetch(`{{ this.url }}`)
        .then(response => response.arrayBuffer())
        .then(arrayBuffer => {
          parseGeoraster(arrayBuffer).then(georaster => {
            var layer = new GeoRasterLayer({
                georaster: georaster,
                opacity: 0.7
            });
            layer.addTo({{this._parent.get_name()}});

            if ("True" == `{{ this.zoom }}`) {
                ({{ this._parent.get_name() }}).fitBounds(layer.getBounds());
            }

        });
        });
    {% endmacro %}
    """  # noqa
    )

    def __init__(self, url, zoom=False):
        super().__init__()
        self.url = url
        self.zoom = zoom


class PrintControl(MacroElement):
    _template = Template(
        """
                {% macro header(this, kwargs) %}
                <link rel="stylesheet" href="https://gitcdn.link/cdn/pasichnykvasyl/Leaflet.BigImage/master/src/Leaflet.BigImage.css">
                <script src='https://unpkg.com/leaflet.bigimage@1.0.1/src/Leaflet.BigImage.js'></script>
                {% endmacro %}
                {% macro script(this, kwargs) %}
                    L.control.BigImage().addTo(({{ this._parent.get_name() }}));
                {% endmacro %}
        """  # noqa
    )



from ecoscope.plotting.plot import (
    EcoPlotData,
    add_seasons,
    collar_event_timeline,
    ecoplot,
    mcp,
    nsd,
    plot_collar_voltage,
    plot_seasonal_dist,
    speed,
)

__all__ = [
    "EcoPlotData",
    "add_seasons",
    "collar_event_timeline",
    "ecoplot",
    "mcp",
    "nsd",
    "speed",
    "plot_collar_voltage",
    "plot_seasonal_dist",
]



import email
import os
import typing

import pandas as pd
import requests
from tqdm.auto import tqdm


def to_hex(val, default="#ff0000"):
    if val and not pd.isnull(val):
        return "#{:02X}{:02X}{:02X}".format(*[int(i) for i in val.split(",")])
    return default


def pack_columns(dataframe: pd.DataFrame, columns: typing.List):
    """This method would add all extra columns to single column"""
    metadata_cols = list(set(dataframe.columns).difference(set(columns)))

    # To prevent additional column from being dropped, name the column metadata (rename it back).
    if metadata_cols:
        dataframe["metadata"] = dataframe[metadata_cols].to_dict(orient="records")
        dataframe.drop(metadata_cols, inplace=True, axis=1)
        dataframe.rename(columns={"metadata": "additional"}, inplace=True)
    return dataframe


def extract_voltage(s: typing.Dict):
    """
    Extracts voltage from different source-provider in EarthRanger
    Parameters
    ----------
    s: typing.Dict

    Returns
    -------
    typing.Any

    """
    additional = s["extra__observation_details"] or {}
    voltage = additional.get("battery", None)  # savannah tracking
    if not voltage:
        voltage = additional.get("mainVoltage", None)  # vectronics
    if not voltage:
        voltage = additional.get("batt", None)  # AWT
    if not voltage:
        voltage = additional.get("power", None)  # Followit
    return voltage


def download_file(url, path, overwrite_existing=False, chunk_size=1024, **request_kwargs):
    r = requests.get(url, stream=True, **request_kwargs)

    if os.path.isdir(path):
        m = email.message.Message()
        m["content-type"] = r.headers["content-disposition"]
        filename = m.get_param("filename")
        if filename is None:
            raise ValueError("URL has no RFC 6266 filename.")
        path = os.path.join(path, filename)

    if os.path.exists(path) and not overwrite_existing:
        print(f"{path} exists. Skipping...")
        return

    with open(path, "wb") as f:
        with tqdm.wrapattr(f, "write", total=int(r.headers["Content-Length"])) as fout:
            for chunk in r.iter_content(chunk_size=chunk_size):
                fout.write(chunk)



import logging

import ee
import numpy as np
import pandas
import pandas as pd
import shapely
import sklearn.mixture
from scipy.stats import norm
from sklearn.preprocessing import LabelEncoder

logger = logging.getLogger(__name__)


def _min_max_scaler(x):
    x = np.array(x, dtype=np.float64)
    x_min = np.nanmin(x)
    x_max = np.nanmax(x)
    x_std = (x - x_min) / (x_max - x_min)  # in the range 0..1
    return x_std


def std_ndvi_vals(aoi=None, start=None, end=None):
    coll = ee.ImageCollection("MODIS/MCD43A4_006_NDVI").select("NDVI").filterDate(start, end)

    ndvi_vals = (
        coll.toBands()
        .reduceRegion("mean", ee.Feature(shapely.geometry.mapping(aoi)).geometry(), bestEffort=True)
        .values()
        .getInfo()
    )
    ndvi_vals_std = _min_max_scaler(ndvi_vals)
    return pd.DataFrame(
        {
            "img_date": pd.to_datetime(coll.aggregate_array("system:time_start").getInfo(), unit="ms", utc=True),
            "NDVI": ndvi_vals_std,
        }
    ).dropna(axis=0)


def val_cuts(vals, num_seasons=2):
    distr = sklearn.mixture.GaussianMixture(n_components=num_seasons, max_iter=500)
    vals = vals["NDVI"].to_numpy().reshape(-1, 1)
    distr.fit(vals)
    mu_vars = np.array(
        sorted(
            zip(distr.means_.flatten(), distr.covariances_.flatten()),
            key=lambda x: x[0],
        )
    )
    cuts = []
    x = np.linspace(0, 1.0, 1000)
    for i in range(1, len(mu_vars)):
        cuts.append(
            np.max(
                x[
                    tuple(
                        [
                            norm.sf(
                                x,
                                loc=mu_vars[tuple([i - 1, 0])],
                                scale=np.sqrt(mu_vars[tuple([i - 1, 1])]),
                            )
                            > norm.cdf(
                                x,
                                loc=mu_vars[tuple([i, 0])],
                                scale=np.sqrt(mu_vars[tuple([i, 1])]),
                            )
                        ]
                    )
                ]
            )
        )
    cuts.append(float("inf"))
    cuts.append(float("-inf"))
    return sorted(cuts)


def seasonal_windows(ndvi_vals, cuts, season_labels):

    enc = LabelEncoder()
    ndvi_vals["season"] = pandas.cut(ndvi_vals["NDVI"], bins=cuts, labels=season_labels)
    ndvi_vals["season_code"] = enc.fit_transform(ndvi_vals["season"])
    ndvi_vals["unique_season"] = (ndvi_vals["season_code"].diff(1) != 0).astype("int").cumsum()
    ndvi_vals["end"] = ndvi_vals["img_date"].shift(-1)

    # Note: There is a slight shift that needs to be done with the date-ranges returned by GEE. What happens is that
    # the image daterange use 12:00 a.m on the given day rather than midnight so it introduces day long gaps in the
    # time windows. Therefore need to add 1 day to each of the end dates to make them align exactly with the next
    # successive start date
    grpd = ndvi_vals.groupby("unique_season")
    return pandas.DataFrame(
        {
            "start": pandas.Series([grp["img_date"].iloc[0] for _, grp in grpd]),
            "end": pandas.Series([grp["img_date"].iloc[-1] for _, grp in grpd]).apply(
                lambda x: x + pd.Timedelta(days=1)
            ),
            "season": pandas.Series(grp["season"].iloc[0] for name, grp in grpd),
            "unique_season": pandas.Series([name for name, _ in grpd]),
        }
    ).set_index("unique_season")



import concurrent.futures
import datetime
import itertools
import math
import re
import typing

import backoff
import geopandas as gpd
import pandas as pd
import requests
from tqdm.auto import tqdm

import ecoscope
from ecoscope.contrib.dasclient import DasClient, DasClientException
from ecoscope.io.utils import pack_columns, to_hex


class BackoffDasClient(DasClient):
    """
    Extends DasClient to add backoff to requests.

    """

    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_tries=10,
    )
    @backoff.on_exception(
        backoff.expo,
        DasClientException,
        max_tries=10,
        giveup=lambda e: len(e.args) == 0
        or not isinstance(e.args[0], str)
        or re.match("Failed to call DAS web service. 50[24] .*", e.args[0]) is None,
    )
    def _get(self, *args, **kwargs):
        return super()._get(*args, **kwargs)


class ConcurrentDasClient(BackoffDasClient):
    """
    Extends DasClient to make multiple requests while maintaining interface.

    """

    def __init__(self, *, sub_page_size=4000, tcp_limit=5, **kwargs):
        self.sub_page_size = sub_page_size
        self.tcp_limit = tcp_limit
        super().__init__(**kwargs)

    def _concurrent_get(self, path, stream=False, **kwargs):
        page_size = kwargs.get("params", {}).get("page_size")

        if page_size is None or page_size % self.sub_page_size or stream:
            return super()._get(path, stream=stream, **kwargs)

        p = kwargs["params"].copy()
        p["page"] = int(p.get("page", 1))
        current_page = p["page"]

        p["page_size"] = 1
        response = super()._get(path=path, params=p)
        p["page_size"] = self.sub_page_size

        if "count" not in response:
            return response

        if not response.get("count"):
            return response

        start = 1 + (p["page"] - 1) * page_size // self.sub_page_size
        end = 1 + min(
            p["page"] * page_size // self.sub_page_size,
            math.ceil(response["count"] / self.sub_page_size),
        )

        if response["previous"] is not None:
            response["previous"] = response["previous"].replace("page_size=1", f"page_size={page_size}")
        if response["count"] > page_size * current_page:
            response["next"] = response["next"].replace("page_size=1", f"page_size={page_size}")
        else:
            response["next"] = None

        with concurrent.futures.ThreadPoolExecutor(max_workers=self.tcp_limit) as executor:
            futures = []
            for page in range(start, end):
                p["page"] = page
                futures.append(executor.submit(super()._get, path=path, params=p.copy()))

            response["results"] = itertools.chain(
                *(future.result()["results"] for future in concurrent.futures.as_completed(futures))
            )
        return response

    def _get(self, *args, **kwargs):
        params = kwargs.get("params", {})
        if "sub_page_size" in params or "tcp_limit" in params:
            print(
                f"Warning: `sub_page_size` and `tcp_limit` should only be provided to the constructor of {type(self)}"
            )
        params["page"] = params.get("page", 1)
        params["page_size"] = params.get("page_size", 1000000000)
        params = {k: v if not isinstance(v, bool) else str(v).lower() for k, v in params.items()}
        kwargs["params"] = params
        return self._concurrent_get(*args, **kwargs)


class EarthRangerIO(ConcurrentDasClient):
    """
    Extends ConcurrentDasClient with Ecoscope-specific configuration.

    """

    def __init__(self, **kwargs):
        if "server" in kwargs:
            server = kwargs.pop("server")
            kwargs["service_root"] = server + "/api/v1.0"
            kwargs["token_url"] = server + "/oauth2/token"

        kwargs["client_id"] = kwargs.get("client_id", "das_web_client")

        super().__init__(**kwargs)

    @staticmethod
    def _to_gdf(df):
        longitude, latitude = (0, 1) if isinstance(df["location"].iat[0], list) else ("longitude", "latitude")
        return gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df["location"].str[longitude], df["location"].str[latitude]),
            crs=4326,
        )

    @staticmethod
    def _normalize_column(df, col):
        print(col)
        for k, v in pd.json_normalize(df.pop(col), sep="__").add_prefix(f"{col}__").iteritems():
            df[k] = v.values

    @staticmethod
    def _clean_kwargs(addl_kwargs={}, **kwargs):
        for k in addl_kwargs.keys():
            print(f"Warning: {k} is a non-standard parameter. Results may be unexpected.")
        return {k: v for k, v in {**addl_kwargs, **kwargs}.items() if v is not None}

    def get_subjectsources(self, subjects=None, sources=None, **addl_kwargs):
        """
        Parameters
        ----------
        subjects
            A comma-delimited list of Subject IDs.
        sources
            A comma-delimited list of Source IDs.

        Returns
        -------
        subjectsources : pd.DataFrame
            DataFrame of queried subjectssources

        """

        kwargs = self._clean_kwargs(
            addl_kwargs,
            subjects=subjects,
            sources=sources,
        )

        return pd.DataFrame(self._get("subjectsources/", params=kwargs)["results"])

    def get_subjects(
        self,
        include_inactive=None,
        tracks_since=None,
        tracks_until=None,
        bbox=None,
        subject_group=None,
        name=None,
        updated_since=None,
        render_last_location=None,
        tracks=None,
        id=None,
        updated_until=None,
        group_name=None,
        **addl_kwargs,
    ):
        """
        Parameters
        ----------
        include_inactive
            Include inactive subjects in list.
        tracks_since
            Include tracks since this timestamp
        tracks_until
            Include tracks up through this timestamp
        bbox
            Include subjects having track data within this bounding box defined by a 4-tuple of coordinates marking
            west, south, east, north.
        subject_group
            Indicate a subject group for which Subjects should be listed.
        name : UUID
            Find subjects with the given name.
        updated_since
            Return Subject that have been updated since the given timestamp.
        render_last_location
            Indicate whether to render each subject's last location.
        tracks
            Indicate whether to render each subject's recent tracks.
        id
            A comma-delimited list of Subject IDs.
        updated_until
        group_name

        Returns
        -------
        subjects : pd.DataFrame
            DataFrame of queried subjects

        """

        kwargs = self._clean_kwargs(
            addl_kwargs,
            include_inactive=include_inactive,
            tracks_since=tracks_since,
            tracks_until=tracks_until,
            bbox=bbox,
            subject_group=subject_group,
            name=name,
            updated_since=updated_since,
            render_last_location=render_last_location,
            tracks=tracks,
            id=id,
            updated_until=updated_until,
            group_name=group_name,
        )

        assert kwargs.get("subject_group") is None or kwargs.get("group_name") is None

        if kwargs.get("group_name") is not None:
            try:
                kwargs["subject_group"] = self._get(
                    "subjectgroups/",
                    params={
                        "group_name": kwargs.pop("group_name"),
                        "include_inactive": True,
                        "include_hidden": True,
                        "flat": True,
                    },
                )[0]["id"]
            except IndexError:
                raise KeyError("`group_name` not found")

        df = pd.DataFrame(self._get("subjects/", params=kwargs)["results"])
        assert not df.empty

        df["hex"] = df["additional"].str["rgb"].map(to_hex) if "additional" in df else "#ff0000"

        return df

    def get_patrols(self, filter=None, status=None, **addl_kwargs):
        """
        Parameters
        ----------
        filter:
            example: {"date_range":{"lower":"2020-09-16T00:00:00.000Z"}}
            date_range
            patrols_overlap_daterange
            text
            patrol_type
            tracked_by
        status
            Comma-separated list of 'scheduled'/'active'/'overdue'/'done'/'cancelled'

        Returns
        -------
        patrols : pd.DataFrame
            DataFrame of queried patrols

        """

        kwargs = self._clean_kwargs(
            addl_kwargs,
            filter=filter,
            status=status,
        )

        return pd.DataFrame(self._get("activity/patrols/", params=kwargs)["results"])

    def get_users(self) -> pd.DataFrame:
        """
        Get all users

        Returns
        -------
        users: pd.DataFrame

        """
        return pd.DataFrame(super().get_users())

    def get_sources(
        self,
        manufacturer_id=None,
        provider_key=None,
        provider=None,
        id=None,
        **addl_kwargs,
    ):
        """
        Parameters
        ----------
        manufacturer_id
        provider_key
        provider
        id

        Returns
        -------
        sources : pd.DataFrame
            DataFrame of queried sources

        """

        kwargs = self._clean_kwargs(
            addl_kwargs,
            manufacturer_id=manufacturer_id,
            provider_key=provider_key,
            provider=provider,
            id=id,
        )

        return pd.DataFrame(self._get("sources/", params=kwargs)["results"])

    def _get_observations(
        self,
        source_ids=None,
        subject_ids=None,
        subjectsource_ids=None,
        tz="UTC",
        since=None,
        until=None,
        filter=None,
        include_details=None,
        created_after=None,
        **addl_kwargs,
    ):
        """
        Return observations matching queries. If `subject_id`, `source_id`, or `subjectsource_id` is specified, the
        index is set to the provided value.

        Parameters
        ----------
        subject_ids
            filter to a single subject
        source_ids
            filter to a single source
        subjectsource_ids
            filter to a subjectsource_id, rather than source_id + time range
        since
            get observations after this ISO8061 date, include timezone
        until
            get observations up to this ISO8061 date, include timezone
        filter
            filter using exclusion_flags for an observation.
            filter=None returns everything
            filter=0  filters out everything but rows with exclusion flag 0 (i.e, passes back clean data)
            filter=1  filters out everything but rows with exclusion flag 1 (i.e, passes back manually filtered data)
            filter=2, filters out everything but rows with exclusion flag 2 (i.e., passes back automatically filtered
            data)
            filter=3, filters out everything but rows with exclusion flag 2 or 1 (i.e., passes back both manual and
            automatically filtered data)
        include_details
            one of [true,false], default is false. This brings back the observation additional field
        created_after
            get observations created (saved in EarthRanger) after this ISO8061 date, include timezone

        Returns
        -------
        observations : gpd.GeoDataFrame
            GeoDataFrame of queried observations

        """

        assert (source_ids, subject_ids, subjectsource_ids).count(None) == 2

        kwargs = self._clean_kwargs(
            addl_kwargs,
            since=since,
            until=until,
            filter=filter,
            include_details=include_details,
            created_after=created_after,
        )

        if source_ids is not None:
            id_name = "source_id"
            ids = source_ids
        elif subject_ids is not None:
            id_name = "subject_id"
            ids = subject_ids
        else:
            id_name = "subjectsource_id"
            ids = subjectsource_ids

        if isinstance(ids, str):
            ids = [ids]

        observations = []
        pbar = tqdm(ids)
        for id in pbar:
            pbar.set_description(f"Downloading Observations for {id_name}={id}")
            kwargs[id_name] = id
            df = pd.DataFrame(self._get("observations/", params=kwargs)["results"])
            df[id_name] = id
            observations.append(df)

        observations = pd.concat(observations)
        if observations.empty:
            return gpd.GeoDataFrame()

        observations["created_at"] = pd.to_datetime(
            observations["created_at"],
            errors="coerce",
            utc=True,
        ).dt.tz_convert(tz)
        observations["recorded_at"] = pd.to_datetime(
            observations["recorded_at"],
            errors="coerce",
            utc=True,
        ).dt.tz_convert(tz)

        observations.sort_values("recorded_at", inplace=True)

        return EarthRangerIO._to_gdf(observations)

    def get_source_observations(self, source_ids, include_source_details=False, relocations=True, **kwargs):
        """
        Get observations for each listed source and create a `Relocations` object.

        Parameters
        ----------
        source_ids : str or list[str]
            List of source UUIDs
        include_source_details : bool, optional
            Whether to merge source info into dataframe
        kwargs
            Additional arguments to pass in the request to EarthRanger. See the docstring of `_get_observations` for
            info.

        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format

        """

        if isinstance(source_ids, str):
            source_ids = [source_ids]

        observations = self._get_observations(source_ids=source_ids, **kwargs)
        if observations.empty:
            return gpd.GeoDataFrame()

        if include_source_details:
            observations = observations.merge(
                pd.DataFrame(self.get_sources(id=",".join(observations["source"].unique()))).add_prefix("source__"),
                left_on="source",
                right_on="source__id",
            )

        if relocations:
            return ecoscope.base.Relocations.from_gdf(
                observations,
                groupby_col="source",
                uuid_col="id",
                time_col="recorded_at",
            )
        else:
            return observations

    def get_subject_observations(
        self,
        subject_ids,
        include_source_details=False,
        include_subject_details=False,
        include_subjectsource_details=False,
        relocations=True,
        **kwargs,
    ):
        """
        Get observations for each listed subject and create a `Relocations` object.

        Parameters
        ----------
        subject_ids : str or list[str]
            List of subject UUIDs
        include_source_details : bool, optional
            Whether to merge source info into dataframe
        include_subject_details : bool, optional
            Whether to merge subject info into dataframe
        include_subjectsource_details : bool, optional
            Whether to merge subjectsource info into dataframe
        kwargs
            Additional arguments to pass in the request to EarthRanger. See the docstring of `__get_observations` for
            info.

        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format

        """

        if isinstance(subject_ids, str):
            subject_ids = [subject_ids]

        observations = self._get_observations(subject_ids=subject_ids, **kwargs)

        if observations.empty:
            return gpd.GeoDataFrame()

        if include_source_details:
            observations = observations.merge(
                pd.DataFrame(self.get_sources(id=",".join(observations["source"].unique()))).add_prefix("source__"),
                left_on="source",
                right_on="source__id",
            )

        if include_subject_details:
            observations = observations.merge(
                self.get_subjects(id=",".join(subject_ids), include_inactive=True).add_prefix("subject__"),
                left_on="subject_id",
                right_on="subject__id",
            )

        if include_subjectsource_details:
            observations = observations.merge(
                self.get_subjectsources(subjects=",".join(observations["subject_id"].unique())).add_prefix(
                    "subjectsource__"
                ),
                left_on=["subject_id", "source"],
                right_on=["subjectsource__subject", "subjectsource__source"],
            )

        if relocations:
            return ecoscope.base.Relocations.from_gdf(
                observations,
                groupby_col="subject_id",
                uuid_col="id",
                time_col="recorded_at",
            )
        else:
            return observations

    def get_subjectsource_observations(
        self,
        subjectsource_ids,
        include_source_details=False,
        relocations=True,
        **kwargs,
    ):
        """
        Get observations for each listed subjectsource and create a `Relocations` object.

        Parameters
        ----------
        subjectsource_ids : str or list[str]
            List of subjectsource UUIDs
        include_source_details : bool, optional
            Whether to merge source info into dataframe
        kwargs
            Additional arguments to pass in the request to EarthRanger. See the docstring of `__get_observations` for
            info.

        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format

        """

        if isinstance(subjectsource_ids, str):
            subjectsource_ids = [subjectsource_ids]

        observations = self._get_observations(subjectsource_ids=subjectsource_ids, **kwargs)

        if observations.empty:
            return gpd.GeoDataFrame()

        if include_source_details:
            observations = observations.merge(
                pd.DataFrame(self.get_sources(id=",".join(observations["source"].unique()))).add_prefix("source__"),
                left_on="source",
                right_on="source__id",
            )

        if relocations:
            return ecoscope.base.Relocations.from_gdf(
                observations,
                groupby_col="subjectsource_id",
                uuid_col="id",
                time_col="recorded_at",
            )
        else:
            return observations

    def get_events(
        self,
        is_collection=None,
        updated_size=None,
        event_ids=None,
        bbox=None,
        include_updates=None,
        include_details=None,
        sort_by=None,
        patrol_segment=None,
        state=None,
        event_type=None,
        include_notes=None,
        include_related_events=None,
        include_files=None,
        max_results=None,
        oldest_update_date=None,
        exclude_contained=None,
        updated_since=None,
        event_category=None,
        filter=None,
        **addl_kwargs,
    ):
        """
        Parameters
        ----------
        is_collection
            true/false whether to filter on is_collection
        updated_since
            date-string to limit on updated_at
        event_ids : array[string]
            Event IDs, comma-separated
        bbox
            bounding box including four coordinate values, comma-separated. Ex. bbox=-122.4,48.4,-122.95,49.0
            (west, south, east, north).
        include_updates
            Boolean value
        include_details
            Boolean value
        sort_by
            Sort by (use 'event_time', 'updated_at', 'created_at', 'serial_number') with optional minus ('-') prefix to
            reverse order.
        patrol_segment
            ID of patrol segment to filter on
        state
            Comma-separated list of 'scheduled'/'active'/'overdue'/'done'/'cancelled'
        event_type
            Comma-separated list of event type uuids
        include_notes
            Boolean value
        include_related_events
            Boolean value
        include_files
            Boolean value
        max_results
        oldest_update_date
        exclude_contained
        event_category
        filter : json dict
            Can contain any of 'event_filter_id', 'text', 'date_range', 'duration', 'state', 'priority',
            'event_category', 'event_type', 'reported_by', 'create_date', 'update_date'

        Returns
        -------
        events : gpd.GeoDataFrame
            GeoDataFrame of queried events

        """

        kwargs = self._clean_kwargs(
            addl_kwargs,
            is_collection=is_collection,
            updated_size=updated_size,
            event_ids=event_ids,
            bbox=bbox,
            include_updates=include_updates,
            include_details=include_details,
            sort_by=sort_by,
            patrol_segment=patrol_segment,
            state=state,
            event_type=event_type,
            include_notes=include_notes,
            include_related_events=include_related_events,
            include_files=include_files,
            max_results=max_results,
            oldest_update_date=oldest_update_date,
            exclude_contained=exclude_contained,
            updated_since=updated_since,
            event_category=event_category,
            filter=filter,
        )

        df = pd.DataFrame(self._get("activity/events/", params=kwargs)["results"])
        assert not df.empty
        df["time"] = pd.to_datetime(df["time"])

        df = gpd.GeoDataFrame(df)
        df.loc[~df["geojson"].isna(), "geometry"] = gpd.GeoDataFrame.from_features(
            df.loc[~df["geojson"].isna(), "geojson"]
        )["geometry"]
        df.set_crs(4326, inplace=True)

        df.sort_values("time", inplace=True)
        return df

    def get_subjectgroup_observations(self, subject_group=None, group_name=None, include_inactive=True, **kwargs):
        """
        Parameters
        ----------
        subject_group : str
            UUID of subject group to filter by
        group_name : str
            Common name of subject group to filter by
        include_inactive : bool, optional
            Whether to get observations for Subjects marked inactive by EarthRanger
        kwargs
            Additional arguments to pass in the request to `get_subject_observations`. See the docstring of
            `get_subject_observations` for info.

        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format

        """

        assert (subject_group is None) != (group_name is None)

        if subject_group:
            subject_ids = self.get_subjects(subject_group=subject_group, include_inactive=include_inactive).id.tolist()
        else:
            subject_ids = self.get_subjects(group_name=group_name, include_inactive=include_inactive).id.tolist()

        return self.get_subject_observations(subject_ids, **kwargs)

    def get_observations_for_patrols(self, patrols_df, **kwargs):
        """
        Download observations for provided `patrols_df`.

        Parameters
        ----------
        patrols_df : pd.DataFrame
            Data returned from a call to `get_patrols`.
        kwargs
            Additional parameters to pass to `get_subject_observations`.

        Returns
        -------
        relocations : ecoscope.base.Relocations

        """

        observations = []
        for _, patrol in patrols_df.iterrows():
            for patrol_segment in patrol["patrol_segments"]:
                subject_id = (patrol_segment.get("leader") or {}).get("id")
                start_time = (patrol_segment.get("time_range") or {}).get("start_time")
                end_time = (patrol_segment.get("time_range") or {}).get("end_time")

                if None in {subject_id, start_time}:
                    continue
                try:
                    observations.append(
                        self.get_subject_observations(
                            subject_ids=[subject_id],
                            since=start_time,
                            until=end_time,
                            **kwargs,
                        )
                    )
                except Exception as e:
                    # print(f'Getting observations for {subject_id=} {start_time=} {end_time=} failed for: {e}')
                    print(
                        f"Getting observations for subject_id={subject_id} start_time={start_time} end_time={end_time}"
                        f"failed for: {e}"
                    )

        for i, patrol in patrols_df.iterrows():
            pd.DataFrame(observations[i])["groupby_col"] = patrol["id"]

        return ecoscope.base.Relocations(pd.concat(observations))

    def post_observations(
        self,
        observations: gpd.GeoDataFrame,
        source_id_col: str = "source",
        recorded_at_col: str = "recorded_at",
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        observations : gpd.GeoDataFrame
            observation data to be uploaded
        source_id_col : str
            The source column in the observation dataframe
        recorded_at_col : str
            The observation recorded time column in the dataframe

        Returns
        -------
        None

        """

        def upload(obs):
            try:
                obs = obs.rename(columns={source_id_col: "source", recorded_at_col: "recorded_at"})
                obs["location"] = pd.DataFrame({"longitude": obs.geometry.x, "latitude": obs.geometry.y}).to_dict(
                    "records"
                )
                del obs["geometry"]
                obs = pack_columns(obs, columns=["source", "recorded_at", "location"])
                post_data = obs.to_dict("records")
                results = super(EarthRangerIO, self).post_observation(post_data)
            except DasClientException as exc:
                self.logger.error(exc)
            except requests.exceptions.RequestException as exc:
                self.logger.error(exc)
            else:
                return pd.DataFrame(results)

        return observations.groupby(source_id_col).progress_apply(upload)

    def post_subjectsource(
        self,
        subject_id: str,
        source_id: str,
        lower_bound_assignend_range: datetime.datetime,
        upper_bound_assigned_range: datetime.datetime,
        additional: typing.Dict = None,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        subject_id
        source_id
        lower_bound_assignend_range
        upper_bound_assigned_range
        additional

        Returns
        -------
        pd.DataFrame

        """

        if additional is None:
            additional = {}
        payload = {
            "source": source_id,
            "assigned_range": {
                "lower": lower_bound_assignend_range,
                "upper": upper_bound_assigned_range,
            },
            "additional": additional,
        }

        urlpath = f"subject/{subject_id}/sources"
        response = self._post(urlpath, payload=payload)
        return pd.DataFrame([response])

    @staticmethod
    def _dataframe_to_dict(events):
        if isinstance(events, gpd.GeoDataFrame):
            events["location"] = pd.DataFrame({"longitude": events.geometry.x, "latitude": events.geometry.y}).to_dict(
                "records"
            )
            del events["geometry"]

        if isinstance(events, pd.DataFrame):
            events = events.to_dict("records")
        return events

    def post_event(
        self,
        events: typing.Union[gpd.GeoDataFrame, pd.DataFrame, typing.Dict, typing.List[typing.Dict]],
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        events

        Returns
        -------
        pd.DataFrame:
            New events created in EarthRanger.

        """

        events = self._dataframe_to_dict(events)
        results = super().post_event(event=events)
        results = results if isinstance(results, list) else [results]
        return pd.DataFrame(results)

    def patch_event(
        self,
        event_id: str,
        events: typing.Union[gpd.GeoDataFrame, pd.DataFrame, typing.Dict, typing.List[typing.Dict]],
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        event_id
            UUID for the event that will be updated.
        events

        Returns
        -------
        pd.DataFrame:
            Updated events in EarthRanger.

        """

        events = self._dataframe_to_dict(events)
        if isinstance(events, list):
            results = [self._patch(f"activity/event/{event_id}", payload=event) for event in events]
        else:
            results = [self._patch(f"activity/event/{event_id}", payload=events)]
        return pd.DataFrame(results)



import concurrent.futures
import datetime as dt
import functools
import itertools
import json
import logging

import backoff
import ee
import numpy as np
import pandas as pd
import pytz
import tqdm.auto as tqdm

logging.getLogger("urllib3").setLevel(logging.ERROR)  # Filter "Connection pool is full, discarding connection"

logger = logging.getLogger(__name__)

ee_is_initialized = False

ee_initialization_expires = dt.datetime(1970, 1, 1, tzinfo=pytz.utc)
ee_initialization_ttl = dt.timedelta(seconds=86400)


# useful stylizations
colourPalettes = {
    "precipitation": {"palette": ["00FFFF", "0000FF"]},
}


def initialize_earthengine(key_dict):
    """
    This takes a JSON key as a dict.
    :param key_dict:
    :return: a credentials object that can be used to initialize the earth engine library.
    """
    global ee_is_initialized
    global ee_initialization_expires

    try:
        if ee_is_initialized and (dt.datetime.now(tz=pytz.utc) > ee_initialization_expires):
            logger.debug("Earth Engine has already been initialized.")
            return

        sac = ee.ServiceAccountCredentials(key_dict["client_email"], key_data=json.dumps(key_dict))
        ee.Initialize(credentials=sac)

    except Exception as e:
        logger.exception("Not able to initialize EarthEngine: %s", e)
        raise
    else:
        ee_is_initialized = True
        ee_initialization_expires = dt.datetime.now(tz=pytz.utc) + ee_initialization_ttl
        logger.debug("Earth Engine has been initialized.")


def convert_millisecs_datetime(unix_time):
    # Define the UNIX Epoch start date
    epoch = dt.datetime(1970, 1, 1, 0, 0, 0)
    t = epoch + dt.timedelta(milliseconds=unix_time)
    return t


def add_img_time(img):
    """
    A function to add the date range of the image as one of its properties and the start and end values as new bands
    """

    # unable to use the band rename() function here in Python whereas works in JS playground
    img = img.addBands(img.metadata("system:time_start"))
    img = img.addBands(img.metadata("system:time_end"))
    dr = ee.DateRange(ee.Date(img.get("system:time_start")), ee.Date(img.get("system:time_end")))
    return img.set({"date_range": dr})


@backoff.on_exception(
    backoff.expo,
    ee.EEException,
    max_tries=10,
)
def label_gdf_with_img(gdf=None, img=None, region_reducer=None, scale=500.0):
    in_fc = ee.FeatureCollection(gdf.__geo_interface__)
    out_fc = img.reduceRegions(in_fc, region_reducer, scale)
    valid_properties = (
        out_fc.first().propertyNames().filter(ee.Filter.inList("item", in_fc.first().propertyNames()).Not())
    )
    return pd.DataFrame(
        out_fc.select(valid_properties)
        .reduceColumns(ee.Reducer.toList(valid_properties.size()), valid_properties)
        .get("list")
        .getInfo(),
        columns=valid_properties.getInfo(),
        index=gdf.index,
    ).apply(pd.Series.explode)


def _match_gdf_to_img_coll_ids(
    gdf, time_col="", img_coll=None, output_col_name=None, stack_limit_before=1, stack_limit_after=1
):
    """
    A function that will add a column to a gdf (output_col_name) that contains
    the stack_limit_before -> stack_limit_after temporally closest image IDs from an image collection.
    :param gdf:
    :param time_col:
    :param img_coll:
    :param output_col_name:
    :param stack_limit_before:
    :param stack_limit_after:
    :return: None
    """

    try:
        # Step 1: download the img_coll image times and ids to a dataframe
        logger.info("Downloading Image Collection IDs and Dates")
        img_data = (
            img_coll.reduceColumns(ee.Reducer.toList(2), ["system:index", "system:time_start"]).get("list").getInfo()
        )
        img_data = np.array(img_data)
        img_data = (
            pd.DataFrame(
                {"img_id": img_data[:, 0], "img_date": pd.to_datetime(img_data[:, 1], unit="ms").tz_localize("UTC")}
            )
            .sort_values("img_date")
            .set_index("img_date")
        )

        # Step 2: determine the closest image IDs to a given feature date
        def determine_img_ids(row):
            row_time = row.get(time_col, pd.Timestamp(0, tz="utc"))
            nearest_index = img_data.index.get_loc(row_time, method="nearest")
            lower = int(nearest_index - stack_limit_before)
            if lower < 0:
                lower = 0
            upper = int(nearest_index + stack_limit_after)
            if upper > len(img_data.index) - 1:
                upper = len(img_data.index)
            img_ids = img_data.iloc[
                lower:upper,
            ]["img_id"].to_list()
            return img_ids

        logger.info("Matching Features to Image IDs")
        gdf[output_col_name] = gdf.apply(determine_img_ids, axis=1)

    except Exception as e:
        logger.error(str(e))


@backoff.on_exception(
    backoff.expo,
    ee.EEException,
    max_tries=10,
)
def label_gdf_with_temporal_image_collection_by_feature(
    gdf=None,
    time_col_name=None,
    stack_limit_before=1,
    stack_limit_after=1,
    img_coll=None,
    region_reducer=None,
    scale=500.0,
):

    # Match the features to the necessary image collection images
    _match_gdf_to_img_coll_ids(
        gdf=gdf,
        time_col=time_col_name,
        img_coll=img_coll,
        output_col_name="img_ids",
        stack_limit_before=stack_limit_before,
        stack_limit_after=stack_limit_after,
    )

    in_fc = ee.FeatureCollection(gdf[["geometry", "img_ids"]].__geo_interface__)

    def feat_func(feat):
        tmp_coll = img_coll.filter(ee.Filter.inList("system:index", feat.get("img_ids")))

        def region_reduc(img):
            return img.set(
                {
                    "img_date": img.date().format(),
                    "img_vals": img.reduceRegion(reducer=region_reducer, geometry=feat.geometry(), scale=scale),
                }
            )

        return feat.set(
            "values",
            tmp_coll.map(region_reduc).reduceColumns(ee.Reducer.toList(2), ["img_date", "img_vals"]).values().get(0),
        )

    logger.info("Labeling Features with Image Collection Values")
    out_fc = in_fc.map(feat_func, True)
    result = pd.DataFrame(
        out_fc.select("values").reduceColumns(ee.Reducer.toList(1), ["values"]).get("list").getInfo(),
        columns=["values"],
        index=gdf.index,
    ).apply(pd.Series.explode)
    result["img_date"] = result["values"].str[0]
    result["values"] = result["values"].str[1]
    result = pd.concat([result.drop(columns=["values"]), result["values"].apply(pd.Series)], ignore_index=False, axis=1)
    return result


@backoff.on_exception(
    backoff.expo,
    ee.EEException,
    max_tries=10,
)
def label_gdf_with_temporal_image_collection_by_timespan(
    gdf=None,
    img_coll=None,
    image_radius=0,
    add_time=False,
    region_reducer="toList",
    df_chunk_size=25000,
    max_workers=1,
):
    img_list = img_coll.toList(img_coll.size())

    # @TODO sort. Current code assumes img_coll is sorted by `system:time_start`
    times = np.array(img_coll.aggregate_array("system:time_start").getInfo())
    time_bins = pd.to_datetime(
        np.concatenate(([0], (times[:-1] + times[1:]) / 2, [pd.Timestamp.max.timestamp() * 1000 - 1])),
        unit="ms",
        utc=True,
    )

    max_idx = len(times)
    imgs, chunks = zip(
        *itertools.chain(
            *[
                [
                    (
                        ee.ImageCollection(
                            img_list.slice(max(idx - image_radius, 0), min(idx + image_radius + 1, max_idx))
                        ).toBands(),
                        gs.iloc[i : i + df_chunk_size],
                    )
                    for i in range(0, len(gs), df_chunk_size)
                ]
                for idx, gs in gdf.geometry.groupby(pd.cut(gdf.fixtime, time_bins, labels=False))
            ]
        )
    )

    def f(img, chunk):
        return label_gdf_with_img(chunk.geometry, img, region_reducer).melt(ignore_index=False)

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        results = list(tqdm.tqdm(executor.map(f, imgs, chunks), total=len(chunks)))

    if results:
        results = pd.concat(results)

        if add_time:
            img_names = np.array(img_coll.aggregate_array("system:index").getInfo())

            cat = results.variable.astype("category").cat
            results["time"] = cat.categories.str[: len(img_names[0])].map(
                pd.Series(pd.to_datetime(times, unit="ms", utc=True), index=img_names)
            )[cat.codes]

        return results


def chunk_gdf(
    gdf=None,
    label_func=None,
    label_func_kwargs=None,
    df_chunk_size=25000,
    max_workers=1,
):
    """
    A function that will process the input gdf in chunks and apply the input label_func function over the chunks.

    :param gdf:
        a geopandas dataframe. The 'geometry' column can be any type pf geometry (point/line/polygon). The gdf needs to
        have a column with the name of the image_collection and the column values are lists of the individual image IDs
        that need to be associated with each feature. This step will typically be run with the
        match_img_coll_ids_to_gdf() function beforehand.
    :param label_func: a function to run on the EE cloud that has the signature (feat, kwargs)
    :param label_func_kwargs: a dictionary of parameters to provide to the label_func
    :param df_chunk_size: how many features (rows) to process at once within EE
    :param max_workers: the number of chunks to process concurrently
    :return: a dataframe with the same index as the input gdf and where each pixel value (or reduced value) is a row
    """
    chunks = [gdf.iloc[i : i + df_chunk_size].copy() for i in range(0, len(gdf), df_chunk_size)]

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        results = list(
            tqdm.tqdm(executor.map(functools.partial(label_func, **label_func_kwargs), chunks), total=len(chunks))
        )

    if results:
        return pd.concat(results)


def calculate_anomaly(
    gdf=None,
    img_coll=None,
    historical_start="2000-01-01",
    start="2010-01-01",
    end="2022-01-01",
    scale=5000.0,
):
    """
    Compute anomalies by subtracting the historical_start mean from each image in a collection of start->end images.
    :param gdf: the input geodataframe
    :param img_coll: the input EE image collection
    :param historical_start: start time for calculating the mean reference
    :param start: end time for mean reference and the start time for the anomaly calculation
    :param end: end time for the anomaly calculation
    :param scale: the image scale
    :return: a dataframe with same index as input gdf with the img_dates and the anomaly calculation
    """
    fc = ee.FeatureCollection(gdf[["geometry"]].__geo_interface__)

    mean_ref = img_coll.filterDate(historical_start, start).mean()

    img_coll = img_coll.filterDate(start, end).sort("system:time_start", False)

    def feat_func(feat):
        def img_func(img):
            return img.set(
                {
                    "img_date": img.date().format(),
                    "mean": img.subtract(mean_ref).reduceRegion("mean", feat.geometry(), scale),
                }
            )

        return feat.set(
            "values",
            img_coll.map(img_func, True).reduceColumns(ee.Reducer.toList(2), ["img_date", "mean"]).values().get(0),
        )

    logger.info("Calculating anomaly")

    vals = fc.map(feat_func, True)
    result = pd.DataFrame(
        vals.select("values").reduceColumns(ee.Reducer.toList(1), ["values"]).get("list").getInfo(),
        columns=["values"],
        index=gdf.index,
    ).apply(pd.Series.explode)

    result["img_date"] = result["values"].str[0]
    result["values"] = result["values"].str[1]
    result.sort_values(by="img_date", inplace=True)
    result = pd.concat(
        [result.drop(columns=["values"]), result["values"].apply(pd.Series).cumsum()], ignore_index=False, axis=1
    )

    return result



import datetime
import typing
from dataclasses import dataclass

import ecoscope


@dataclass
class ImmobilityProfile:
    """
    threshold_radius: radius of cluster in metres.  Defaults to 13m.
    threshold_time: time in seconds the track is expected to be stationary. Defaults to 18000 seconds (5 hours)
    threshold_probability: the proportion of observations in a sample which
    must be inside a cluster to generate a CRITICAL.  Default 0.8

    """

    threshold_radius: int = 13
    threshold_time: int = 18000
    threshold_probability: float = 0.8


class Immobility:
    @classmethod
    def calculate_immobility(
        cls, immobility_profile: ImmobilityProfile, relocs: ecoscope.base.Relocations
    ) -> typing.Dict:
        """
        Function to search for immobility within a movement trajectory. Assumes we start with a filtered
        trajectory spanning some period of time. The algorithm will work backwards through the trajectory's
        relocations and build a cluster. Looks to see if the cluster characteristics match immobility criteria
        (ie., timespan is gte than the threshold_time, and the cluster probability is gte to the threshold_probability)
        Note that this is a simplified version of the full clustering algorithm since it's only looking at data within
        threshold time and will not figure out the true start of an immobility without looking backwards through all
        possible points
        TODO: include more info about the immobility result:
        1) immobility start time
        2) immobility probability
        3) immobility cluster fix count
        4) algorithm provenance

        Parameters
        ----------
        immobility_profile: ImmobilityProfile
            setting for immobility
        relocs: ecoscope.base.Relocations
            set for fixes for given subject.

        Returns
        -------
            typing.Dict

        """

        relocs.remove_filtered(inplace=True)
        relocs.sort_values(by="fixtime", ascending=True, inplace=True)
        ts = relocs.fixtime.iat[-1] - relocs.fixtime.iat[0]

        if ts < datetime.timedelta(seconds=immobility_profile.threshold_time):
            raise Exception("Insufficient Data")

        relocs.sort_values(by="fixtime", ascending=False, inplace=True)

        cluster_pvalue = (
            relocs.threshold_point_count(threshold_dist=immobility_profile.threshold_radius) / relocs.shape[0]
        )

        if (cluster_pvalue >= immobility_profile.threshold_probability) and (
            ts.total_seconds() > immobility_profile.threshold_time
        ):
            return {
                "probability_value": cluster_pvalue,
                "cluster_radius": relocs.cluster_radius,
                "cluster_fix_count": relocs.threshold_point_count(immobility_profile.threshold_radius),
                "total_fix_count": relocs.shape[0],
                "immobility_time": ts,
            }



import logging
import math
import os
from collections import UserDict

import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import rasterio as rio
import rasterio.mask
import tqdm.auto as tqdm

logger = logging.getLogger(__name__)


class RasterExtent:
    # minx, miny, maxx, maxy
    def __init__(self, x_min=33.0, x_max=37.0, y_min=2.0, y_max=-2.0):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max

    def __repr__(self):
        return "RasterExtent(xmin: {0}, xmax: {1}, ymin: {2}, ymax: {3})".format(
            self.x_min, self.x_max, self.y_min, self.y_max
        )

    @classmethod
    def create_from_origin(cls, pixel_size=0.5, x_min=0.0, y_min=0.0, num_rows=100, num_columns=100):
        max_x = x_min + (num_columns * pixel_size)
        max_y = y_min + (num_rows * pixel_size)
        return cls(x_min=x_min, x_max=max_x, y_min=y_min, y_max=max_y)


class RasterProfile(UserDict):
    """
    A class for holding raster properties
    At present this class is only valid for non-rotated rasters with a north-up orientation and square sized pixels
    defined by the E-W pixel size
    """

    def __init__(
        self,
        pixel_size=1000.0,
        pixel_dtype=rio.float64,
        crs="EPSG:8857",  # WGS 84/Equal Earth Greenwich
        nodata_value=0.0,
        band_count=1,
        raster_extent=None,
    ):
        crs = pyproj.CRS.from_user_input(crs)
        raster_extent = raster_extent or RasterExtent()
        cols = int(math.ceil((raster_extent.x_max - raster_extent.x_min) / pixel_size))
        rows = int(math.ceil((raster_extent.y_max - raster_extent.y_min) / pixel_size))
        affine_transform = rio.Affine(pixel_size, 0.0, raster_extent.x_min, 0.0, -pixel_size, raster_extent.y_max)
        super().__init__(
            self,
            crs=crs,
            pixel_size=pixel_size,
            nodata_value=nodata_value,
            columns=cols,
            rows=rows,
            dtype=pixel_dtype,
            band_count=band_count,
            raster_extent=raster_extent,
            transform=affine_transform,
        )

    def _recompute_transform_(self, key):
        """Recomputes the affine transformation matrix when the pixel_size or raster_extent value is updated."""
        _names = ["pixel_size", "raster_extent"]
        if all([key in _names, hasattr(self, _names[0]), hasattr(self, _names[1])]):
            xmin = self.raster_extent.x_min
            ymax = self.raster_extent.y_max
            pixel_size = self.pixel_size

            transform = rio.Affine(pixel_size, 0.0, xmin, 0.0, -pixel_size, ymax)
            cols = int(math.ceil((self.raster_extent.x_max - self.raster_extent.x_min) / self.pixel_size))
            rows = int(math.ceil((self.raster_extent.y_max - self.raster_extent.y_min) / self.pixel_size))

            self.update(transform=transform, columns=cols, rows=rows)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name == "data":
            object.__setattr__(self, name, value)
        else:
            self[name] = value
            self._recompute_transform_(name)

    def __setitem__(self, key, item):
        self.data[key] = item
        self._recompute_transform_(key)


class RasterPy:
    @classmethod
    def write(
        cls,
        ndarray,
        fp,
        columns,
        rows,
        band_count,
        driver="GTiff",
        dtype=rio.float64,
        crs=None,
        transform=None,
        nodata=None,
        sharing=False,
        indexes=1,
        **kwargs,
    ):
        with rio.open(
            fp=fp,
            mode="w",
            driver=driver,
            height=rows,
            width=columns,
            count=band_count,
            dtype=dtype,
            crs=crs,
            nodata=nodata,
            sharing=sharing,
            transform=transform,
        ) as r:
            r.write(ndarray, indexes=indexes)

    @classmethod
    def read(cls, fp, driver=None, **kwargs):
        fp = os.fspath(fp)
        path = rio.parse_path(fp)
        return rio.DatasetReader(path, driver=driver, **kwargs)

    @classmethod
    def read_write(cls, fp, driver=None, **kwargs):
        fp = os.fspath(fp)
        path = rio.parse_path(fp)
        return rio.get_writer_for_path(path, driver=driver)(path, "r+", driver=driver, **kwargs)


def reduce_region(gdf, raster_path_list, reduce_func):
    """
    A function to apply the reduce_func to the values of the pixels within each of the rasters for every
    shape within the input geopandas dataframe 'geometry' column
    :param gdf: geopandas dataframe. The geometry column will be used to mask the areas of the input
    raster to be used
    in the reduction
    :param raster_path_list: a list of raster files on disc to analyse
    :param reduce_func: a single-value function to apply to the values of the input raster
    :return: dataframe with a column of reduce values for each raster and a row for each region
    """

    reduce_func = {
        np.mean: np.nanmean,
        np.sum: np.nansum,
        np.min: np.nanmin,
        np.max: np.nanmax,
    }.get(reduce_func, reduce_func)

    d = {}
    for raster_path in tqdm.tqdm(raster_path_list):
        d[raster_path] = {}
        with rio.open(raster_path) as src:
            for i, shp in gdf.geometry.to_crs(src.crs).iteritems():
                try:
                    d[raster_path][i] = reduce_func(rio.mask.mask(src, [shp], filled=False)[0].compressed())
                except ValueError as e:
                    logger.exception(raster_path, i, e)

    return pd.DataFrame(d)


def raster_to_gdf(raster_path):
    with rio.open(raster_path) as src:
        image = src.read(1)

        dtype = {
            np.float64: np.float32,
            np.int64: np.int32,
            np.uint64: np.int32,
            np.uint32: np.int32,
        }.get(image.dtype.type, image.dtype)

        if dtype is not image.dtype:
            print(f"Error: {image.dtype} is not supported. Converting to {dtype}. Data may be lost.")
            image = image.astype(dtype)

        mask = src.dataset_mask()
        return gpd.GeoDataFrame.from_features(
            [
                {"properties": {"value": value}, "geometry": polygon}
                for polygon, value in rio.features.shapes(image, transform=src.transform, mask=mask)
                if pd.notna(value)
            ],
            crs=src.crs,
        )



from ecoscope.base._dataclasses import (
    RelocsCoordinateFilter,
    RelocsDateRangeFilter,
    RelocsDistFilter,
    RelocsSpeedFilter,
    TrajSegFilter,
)
from ecoscope.base.base import EcoDataFrame, Relocations, Trajectory
from ecoscope.base.utils import (
    cachedproperty,
    create_meshgrid,
    groupby_intervals,
)

__all__ = [
    "EcoDataFrame",
    "Relocations",
    "RelocsCoordinateFilter",
    "RelocsDateRangeFilter",
    "RelocsDistFilter",
    "RelocsSpeedFilter",
    "TrajSegFilter",
    "Trajectory",
    "cachedproperty",
    "create_meshgrid",
    "groupby_intervals",
]
